import busio
import board
from reactor_code_framework.analog_to_digital_converter.generic_adc import GenericADC

import adafruit_ads1x15
from adafruit_ads1x15.ads1015 import ADS1015
from adafruit_ads1x15.ads1115 import ADS1115
from adafruit_ads1x15.analog_in import AnalogIn

#NOTE: the installation instructions for CircuitPython bundle (which includes
#adafruit_ads1x15) can be found at the link below.
#https://learn.adafruit.com/circuitpython-on-raspberrypi-linux/installing-circuitpython-on-raspberry-pi
#https://docs.circuitpython.org/projects/ads1x15/en/latest/

#The old adafruit_ads1x15 ADC module (and examples) (used on Ivysaur/Bulbasaur):
#https://github.com/adafruit/Adafruit_Python_ADS1x15
#https://github.com/adafruit/Adafruit_Python_ADS1x15/blob/master/examples/differential.py
#https://docs.circuitpython.org/projects/ads1x15/en/0.5.2/api.html

#Adafruit CircuitPython bundle on github:
#https://github.com/adafruit/Adafruit_CircuitPython_Bundle
#https://github.com/adafruit/circuitpython

#Adafruit CircuitPython ADS1x15 documentation (new):
#https://cdn-learn.adafruit.com/downloads/pdf/adafruit-4-channel-adc-breakouts.pdf
#https://docs.circuitpython.org/projects/ads1x15/en/latest/api.html?highlight=AnalogIn#adafruit_ads1x15.analog_in.AnalogIn
#https://github.com/adafruit/Adafruit_CircuitPython_ADS1x15

#ADS1x15 Examples/Walkthroughs:
#https://learn.adafruit.com/raspberry-pi-analog-to-digital-converters/ads1015-slash-ads1115
#https://learn.adafruit.com/adafruit-4-channel-adc-breakouts/python-circuitpython
#https://learn.adafruit.com/circuitpython-essentials/circuitpython-analog-in

#ADS1x15 datasheet (technical details):
#https://cdn-shop.adafruit.com/datasheets/ads1015.pdf

#This class implements the functionality for both AdafruitADS1015 and
#AdafruitADS1115.
class AdafruitADS1x15(GenericADC):
  #Given a ReactorManager instance (reactor_manager), a string representing the
  #class (adc_type), and an instance of the appropriate ads class based on the
  #adc_type (ads_instance), a dictionary mapping pin strings to the underlying
  #values (pin_reference), the i2c address, a number representing one of the 
  #gain settings, and the number of readings to average
  def __init__(self, reactor_manager, adc_type, ads_instance, pin_reference, 
               i2c_address, gain = 1, num_readings_to_average = 5):
    self.ads_instance = ads_instance
    self.valid_gains = [2/3, 1, 2, 4, 8, 16]
    self.i2c_address = i2c_address
    #This can be used to get the underlying value used for the pin.
    self.pin_reference = pin_reference
    #These variables are used to store the channels that have already
    #been created.
    self.single_ended_channels = dict()
    self.differential_channels = dict()
    super().__init__(reactor_manager, adc_type, ("A0", "A1", "A2", "A3"),
                     ("A0", "A1", "A2", "A3"), 
                     (("A0", "A1"), ("A0", "A3"), ("A1", "A3"), ("A2", "A3")), 
                     gain, "I2C:{0}".format(self.i2c_address), 
                     num_readings_to_average = num_readings_to_average)

  #Check that the gain setting is valid.
  def check_parameters(self):
    super().check_parameters()
    if self.gain not in self.valid_gains:
      raise Exception("{0} Error({1}): Gain Setting {2} is not one of the valid settings ({3}).".format(self.adc_type, self, self.gain, self.valid_gains))
  
  #Given a gain setting (gain), store the gain and set the gain. 
  def set_gain(self, gain):
    if gain not in self.valid_gains:
      raise Exception("{0} Error({1}): Gain Setting {2} is not one of the valid settings ({3}).".format(self.adc_type, self, gain, self.valid_gains))
    self.gain = gain
    self.ads_instance.gain = gain
  
  # Given a pin (positive_pin) and possibly a negative/reference pin 
  # (negative_pin), and a string equal to "value" or "voltage" (mode), read the 
  # value or voltage (whichever is specified) of:
  # 1) the pin (if negative_pin is None)
  # 2) the pin minus the reference pin (if negative_pin is not None)
  #
  # NOTE: the valid pins are "A0", "A1", "A2", and "A3".
  #
  # NOTE: this function is not meant to be called by the user; use "read"
  # instead, which averages multiple readings.
  def read_single_measurement(self, positive_pin, negative_pin = None, 
                              mode = "voltage"):
    # Single point reading.
    if negative_pin == None:
      channel = self.single_ended_channels.get(positive_pin, None)
      if channel == None:
        channel = AnalogIn(self.ads_instance, self.pin_reference[positive_pin])
        self.single_ended_channels[positive_pin] = channel
    # Differential reading.
    else:
      channel = self.differential_channels.get((positive_pin, negative_pin), 
                                               None)
      if channel == None:
        channel = AnalogIn(self.ads_instance, self.pin_reference[positive_pin],
                           self.pin_reference[negative_pin])
        self.differential_channels[(positive_pin, negative_pin)] = channel

    # Return the reading.
    if mode == "voltage":
      return channel.voltage
    elif mode == "value":
      return channel.value
  
  #No cleanup is needed for this hardware.
  def cleanup(self):
    return

#This class is meant to be used with the ADS1015 12-Bit ADC - 4 Channel.
#Link to product: https://www.adafruit.com/product/1083
class AdafruitADS1015(AdafruitADS1x15):
  def __init__(self, reactor_manager, i2c_address, gain = 1, 
               num_readings_to_average = 5):
    i2c = busio.I2C(board.SCL, board.SDA)
    ads_instance = ADS1015(i2c, address = i2c_address, gain = gain)
    pin_reference = {"A0": adafruit_ads1x15.ads1015.P0, 
                     "A1": adafruit_ads1x15.ads1015.P1,
                     "A2": adafruit_ads1x15.ads1015.P2,
                     "A3": adafruit_ads1x15.ads1015.P3}
    super().__init__(reactor_manager, "AdafruitADS1015", ads_instance, 
                     pin_reference, i2c_address, gain = gain, 
                     num_readings_to_average = num_readings_to_average)

#This class is meant to be used with the ADS1115 16-Bit ADC - 4 Channel.
#Link to the product: https://www.adafruit.com/product/1085
class AdafruitADS1115(AdafruitADS1x15):
  def __init__(self, reactor_manager, i2c_address, gain = 1, 
               num_readings_to_average = 5):
    i2c = busio.I2C(board.SCL, board.SDA)
    ads_instance = ADS1015(i2c, address = i2c_address, gain = gain)
    pin_reference = {"A0": adafruit_ads1x15.ads1115.P0, 
                     "A1": adafruit_ads1x15.ads1115.P1,
                     "A2": adafruit_ads1x15.ads1115.P2,
                     "A3": adafruit_ads1x15.ads1115.P3}
    super().__init__(reactor_manager, "AdafruitADS1115", ads_instance, 
                     pin_reference, i2c_address, gain = gain, 
                     num_readings_to_average = num_readings_to_average)