def generate_generic_adc_error_string(function_name):
  return "Generic ADC Instance: DO NOT USE! " + \
         "Please override {0} in the ".format(function_name) + \
         "specific adc type's sub class."

#This is the generic class for analog-to-digital converters. 
#NOTE: This class should not be used. Instead, you should use one of its 
#subclasses.
class GenericADC(object):
  #Given a ReactorManager instance (reactor_manager), a string representing the
  #type of adc (adc_type), a tuple containing the pins associated with the
  #adc (adc_pins), a tuple which contains the pins that are valid pins for 
  #single-ended reading (valid_single_ended_pins), a tuple which contains pairs
  #of pins (as tuples), which are valid for differential reading, and a variable 
  #which represents the gain setting (gain), a variable that contains any 
  #additional information about how the adc is connected to the raspberry pi 
  #(position) (for example, i2c address), and a variable that specifies the
  #number of readings to average.
  #NOTE: this class should not be instantiated directly. Instead, use one of
  #the subclasses.
  def __init__(self, reactor_manager, adc_type, adc_pins, 
               valid_single_ended_pins, valid_differential_pins, gain, 
               position, num_readings_to_average = 5):
    self.reactor_manager = reactor_manager
    self.adc_type = adc_type
    self.adc_pins = adc_pins
    self.valid_single_ended_pins = valid_single_ended_pins
    self.valid_differential_pins = valid_differential_pins
    self.gain = gain
    self.num_readings_to_average = num_readings_to_average
    if self.gain != None:
      self.set_gain(gain)
    self.position = position
    self.check_parameters()
    #Pass the instance for the device to the reactor manager.
    self.reactor_manager.add_hardware_device(self)

  #Check that the input parameters are valid. If there are additional parameters
  #to check in the subclass, this function can be overridden (make sure
  #to call super().check_parameters() in the overridden version).
  def check_parameters(self):
    if not isinstance(self.adc_pins, tuple):
      raise Exception("{0} Error({1}): ADC Pins {2} is not a tuple.".format(self.adc_type, self, self.adc_pins))
    for pin in self.valid_single_ended_pins:
      if pin not in self.adc_pins:
        raise Exception("{0} Error({1}): Valid Single Ended ADC Pin {2} not in ADC Pins {3}.".format(self.adc_type, self, pin, self.adc_pins))
    for (pos_pin, neg_pin) in self.valid_differential_pins:
      if pos_pin not in self.adc_pins:
        raise Exception("{0} Error({1}): Differential Positive ADC Pin {2} not in ADC Pins {3}.".format(self.adc_type, self, pos_pin, self.adc_pins))
      if neg_pin not in self.adc_pins:
        raise Exception("{0} Error({1}): Differential Negative ADC Pin {2} not in ADC Pins {3}.".format(self.adc_type, self, neg_pin, self.adc_pins))
  
  #Return a string representing the adc.
  def __repr__(self):
    return "{0}: ADC Pins: {1}  Gain: {2}  Position: {3}".format(self.adc_type, self.adc_pins, self.gain, self.position)

  #Allow for equality comparison between two adc.
  def __eq__(self, other):
    return self.adc_type == other.adc_type and \
           self.adc_pins == other.adc_pins and \
           self.gain == other.gain and \
           self.position == other.position
  
  #Allow for ADC instances to be used in sets and as keys in dictionaries.
  def __hash__(self):
    return hash(self.adc_type, self.adc_pins, self.gain, self.position)

  #Using the gain setting stored in self.gain, set the gain of the adc.
  def set_gain(self):
    raise Exception(generate_generic_adc_error_string("set_gain"))
  
  #This function is used to check the pins passed into the read method if they
  #are valid. It references the variables self.valid_single_ended_pins and
  #self.valid_differential_pins, and uses these to determine if the pin)s) is
  #(are) valid.
  #NOTE: if negative_pin == None, this function checks positive_pin against
  #self.valid_single_ended_pins. Otherwise, this functions checks 
  #the pair of positive and negative pins against self.valid_differential_pins.
  def check_pin_validity(self, positive_pin, negative_pin = None):
    if negative_pin == None:
      if positive_pin not in self.valid_single_ended_pins:
        raise Exception("{0} Error({1}): Pin {2} is not one of the valid pins for single-ended reading (valid pins: {3}).".format(self.adc_type, self, positive_pin, self.valid_single_ended_pins))
    else:
      if (positive_pin, negative_pin) not in self.valid_differential_pins:
        raise Exception("{0} Error({1}): Positive Pin-Negative Pin pair {2} is not one of the valid pairs for differential reading (valid pin pairs: {3}).".format(self.adc_type, self, (positive_pin, negative_pin), self.valid_differential_pins))
  
  #Checks that the mode is valid (either "value" or "voltage").
  def check_mode(self, mode):
    if not (mode == "value" or mode == "voltage"):
      raise Exception("{0} Error({1}): Mode {2} is not 'value' or 'voltage'.".format(self.adc_type, self, mode))
  
  # Given a pin (positive_pin) and possibly a negative/reference pin 
  # (negative_pin), and a string equal to "value" or "voltage" (mode), read the 
  # value or voltage (whichever is specified) of:
  #
  #   1) the pin (if negative_pin is None)
  #   2) the pin minus the reference pin (if negative_pin is not None)
  #
  # NOTE: this function is not meant to be called by the user; use "read"
  # instead, which averages multiple readings.
  def read_single_measurement(self, positive_pin, negative_pin = None, 
                              mode = "voltage"):
    raise Exception(generate_generic_adc_error_string("read_single_measurement"))
  
  def read(self, positive_pin, negative_pin = None, mode = "voltage"):
    self.check_pin_validity(positive_pin, negative_pin)
    self.check_mode(mode)

    sum_of_readings = 0
    for i in range(self.num_readings_to_average):
      reading = self.read_single_measurement(positive_pin, 
                                             negative_pin = negative_pin, 
                                             mode = mode)
      sum_of_readings += reading
    
    return sum_of_readings / self.num_readings_to_average

  #Perform any cleanup for the hardware.
  def cleanup(self):
    raise Exception(generate_generic_adc_error_string("cleanup"))