import time
import os

def generate_generic_camera_error_string(function_name):
  return "Generic Camera Instance: DO NOT USE! " + \
         "Please override {0} in the ".format(function_name) + \
         "specific camera type's sub class."

#Generic Camera Class. Not intended for use in projects. Use a subclass
#instead.
class GenericCamera(object):
  #Given a ReactorManager (reactor_manager), a string representing the camera 
  #type, the directory that the data will be saved in (save_dir), the 
  #resolution, the rotation (0 to do nothing or 180 to turn upside down), the 
  #reflection (a tuple of booleans; if the first is True, the image will be 
  #flipped horizontally, and if the second is True, the image will be flipped 
  #vertically), and any variables associated with the position (think MultiCam, 
  #how each camera is differentiated) store them. Create the save directory if 
  #it doesn't exist.
  #
  #NOTE: resolution is width x height (i.e. (width, height)).
  #
  #NOTE: if resolution, rotation, or reflection is passed as None, then the 
  #camera will not set a specific resolution, rotation, or reflection. Instead, 
  #the camera will attempt to use the default resolution/rotation/reflection.
  def __init__(self, reactor_manager, camera_type, save_dir, resolution = None, 
               rotation = None, reflection = None, position = None):
    self.reactor_manager = reactor_manager
    self.camera_type = camera_type
    self.save_dir = save_dir

    self.resolution = resolution
    self.rotation = rotation
    self.reflection = reflection

    self.subfolders = []
    self.position = position

    self.check_parameters()

    if resolution != None:
      self.resolution = tuple(self.resolution)
      self.set_initial_resolution(self.resolution)
    
    if rotation != None:
      self.set_initial_rotation(self.rotation)
    
    if reflection != None:
      self.reflection = tuple(self.reflection)
      self.set_initial_reflection(self.reflection)
    
    #Add the camera to the reactor manager.
    self.reactor_manager.add_hardware_device(self)
    #Make the directory for where the data is to be saved.
    self.reactor_manager.make_directory(save_dir)

  #Raise an exception if the rotation value isn't valid (build off of in sub
  #class).
  def check_parameters(self):
    if not (self.rotation == None or self.rotation in [0, 90, 180, 270]):
      raise Exception("{0} Error({1}): Rotation {2} is not equal to 0, 90, 180, 270.".format(self.camera_type, self, self.rotation))
  
  #Return a string representing the camera (build off of in sub class).
  def __repr__(self):
    return "{0}: Resolution: {1}  Rotation: {2}  Reflection: {3}  Save Directory: {3}  Additional Position Info: {4}".format(self.camera_type, self.resolution, self.rotation, self.reflection, self.save_dir, self.position)
  
  #Set the resolution of the camera, using the initial resolution (implement
  #in sub class).
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE RESOLUTION MID RUN.
  def set_initial_resolution(self, resolution):
    raise Exception(generate_generic_camera_error_string("set_initial_resolution")) 

  #Set the rotation of the camera, using the inital rotation (implement in
  #sub class).
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE ROTATION MID RUN.
  def set_initial_rotation(self, rotation):
    raise Exception(generate_generic_camera_error_string("set_initial_rotation"))
  
  #Set the reflection of the camera, using the inital reflection (implement in
  #sub class).
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE REFLECTIOn MID RUN.
  def set_initial_reflection(self, reflection):
    raise Exception(generate_generic_camera_error_string("set_initial_reflection"))
  
  #Set the auto white balance of the camera (implement in sub class); only works
  #for Raspberry Pi Camera.
  def set_auto_white_balance(self, auto_white_balance):
    raise Exception(generate_generic_camera_error_string("set_auto_white_balance"))

  #Allow for the comparison of two cameras.
  def __eq__(self, other):
    return self.camera_type == other.camera_type and \
           self.save_dir == other.save_dir and \
           self.position == other.position

  #Hash the camera type.
  def __hash__(self):
    return hash(self.camera_type, self.save_dir, self.position)
  
  #Return the directory that the pictures are saved to.
  def get_save_dir(self):
    return self.save_dir
  
  #This function actually captures and saves the picture (implement in sub
  #class).
  #NOTE: to take pictures, do not use this function, use take_picture.
  #This function is for internal use only.
  def capture_picture(self, picture_path, banner_data):
    raise Exception(generate_generic_camera_error_string("capture_picture"))

  #Take a picture.
  #If save_with_time is True, use the time to name the picture. The default
  # is True.
  #
  #If file_name_prefix is not equal to None, use it as a prefix for the file
  #name of the picture. The default is None.
  #
  #Give the file extension "jpg" or "png". The default is "png".
  #
  #Save the meta_data in a yaml file (NOTE: UNIMPLEMENTED). The default is None.
  #
  #If subfolder_name is not None, the sub-folder specified by subfolder_name 
  #will be created, and the picture will be saved in that subfolder. If it is
  #None, the picture will be saved in the folder dictated by self.save_dir. The
  #default is None.
  #
  #banner_data IS ONLY IMPLEMENTED FOR USB CAMERAS. If it is not None, it
  #should be a list of three strings, where the first string is the "title", 
  #the second string is the "sub-title", and the third string is "info". Then,
  #a banner with this data will be placed at the bottom of the image taken by
  #the camera. Otherwise, if banner_data is None, no banner will be drawn.
  #
  #If auto_white_balance is None, no auto white balancing occurs. Otherwise,
  #pass in a tuple of (red_gain, blue_gain).
  #
  #NOTE: Either save_with_time must be True or file_name_prefix != None.
  #
  #Return the file path to the picture.
  def take_picture(self, save_with_time = True, file_name_prefix = None, 
                   file_extension = "png", meta_data = None, 
                   subfolder_name = None, banner_data = None, 
                   auto_white_balance = None):
    #Check the disk space.
    self.reactor_manager.check_disk_space()

    if ((not save_with_time) and (file_name_prefix == None)):
      raise Exception("Picture Taking Error: Either you must save with the time or the file name prefix must not be None.")
    
    picture_file_name = ""
    #Add the file name prefix.
    if file_name_prefix != None:
      picture_file_name = picture_file_name + file_name_prefix
    
    #Add the time to the file name, if indicated.
    if save_with_time:
      date_from_time = time.localtime(time.time())
      picture_file_name = picture_file_name + \
        time.strftime("y%ym%md%dH%HM%MS%S", date_from_time)
    
    if auto_white_balance != None:
      self.set_auto_white_balance(auto_white_balance)

    #Create the subfolder, if necessary.
    if subfolder_name != None:
      picture_directory = os.path.join(self.save_dir, subfolder_name)
      if subfolder_name not in self.subfolders:
        self.reactor_manager.make_directory(picture_directory)
        self.subfolders.append(subfolder_name)
    else:
      picture_directory = self.save_dir

    picture_path = os.path.join(picture_directory, 
                                picture_file_name  + "." + file_extension)
    
    self.capture_picture(picture_path, banner_data)
    #DEAL WITH METADATA HERE
    return picture_path

  #Run any code to cleanup the camera at the end (implement in sub class).
  def cleanup(self):
    raise Exception(generate_generic_camera_error_string("cleanup"))