from reactor_code_framework.camera.generic_camera import GenericCamera

import RPi.GPIO as GPIO

#Installation instructions for the ArduCam Multi Camera Adapter:
#https://github.com/ArduCAM/RaspberryPi/tree/master/Multi_Camera_Adapter/Multi_Adapter_Board_4Channel

#Another library for working with this hardware:
#https://www.timpoulsen.com/2018/multiple-cameras-with-a-single-raspberry-pi.html#multiple-cameras-with-a-single-raspberry-pi
#https://gist.github.com/skypanther/04b44eaa455f358eb70e59336c068312#file-multicam-py-L154

#The less helpful, but official, documentation:
#https://www.uctronics.com/download/Amazon/B0120.pdf
#Older version: https://www.arducam.com/downloads/RaspCAM/RaspberryPi_Multi_Camera_Adapter_Module_UG.pdf

#This class is intended to be used for the multiple camera attachment for the
#raspberry pi cameras.
#NOTE: in the past, our USB cameras have taken pictures with a resolution of
#2592x1944, but you can feel free to use other resolutions.
#NOTE: Passing different resolutions to each camera has not been tested and
#may lead to odd behavior. Check that everything is working if you try this.
#NOTE: The following is an example for taking pictures with MultiCamera.
#   multi_cam = MultiCamera(reactor_manager, ...)
#   multi_cam.camera_A.take_picture(...)
# OR
#   multi_cam = MultiCamera(reactor_manager, ...)
#   camera_A = multi_cam.get_camera("A")
#   camera_A.take_picture(...)
#NOTE: some aspects of this class (mostly the pin naming and setting
#conventions) come from 
#https://gist.github.com/skypanther/04b44eaa455f358eb70e59336c068312#file-multicam-py-L154
class MultiCamera(object):
  #Take in a ReactorManager instance (reactor_manager), a list of the active 
  #cameras names, a dictionary of save directories (mapping active cameras to 
  #the save directories) a dictionary of the resolutions (mapping active cameras 
  #to resolutions), a dictionary of rotations (mapping active cameras to 
  #rotations), and a dictionary of reflections (mapping active cameras to 
  #reflections).
  #
  #NOTE: the only allowed names for cameras are "A", "B", "C", "D".
  #
  #NOTE: if resolution_dict, rotation_dict, or reflection_dict is passed as 
  #None, then each split camera will not set a specific resolution, rotation, or 
  #reflection. Instead, each split camera will attempt to use the default 
  #resolution/rotation/reflection.
  def __init__(self, reactor_manager, active_camera_names, save_dir_dict, 
               resolution_dict = None, rotation_dict = None, 
               reflection_dict = None):
    #Store and check the user-defined parameters.
    self.reactor_manager = reactor_manager
    self.active_camera_names = active_camera_names
    self.save_dir_dict = save_dir_dict

    #Create the resolution, rotation, and/or reflection dictionaries if
    #necessary.
    if resolution_dict == None:
      self.resolution_dict = dict()
      for camera_name in self.active_camera_names:
        self.resolution_dict[camera_name] = None
    else:
      self.resolution_dict = resolution_dict

    if rotation_dict == None:
      self.rotation_dict = dict()
      for camera_name in self.active_camera_names:
        self.rotation_dict[camera_name] = None
    else:
      self.rotation_dict = rotation_dict
    
    if reflection_dict == None:
      self.reflection_dict = dict()
      for camera_name in self.active_camera_names:
        self.reflection_dict[camera_name] = None
    else:
      self.reflection_dict = reflection_dict

    self.check_parameters()

    #Add the MultiCamera instance to the hardware manager.
    self.reactor_manager.add_hardware_device(self)

    #Instantiate the cameras.
    self.cameras = dict()
    for camera_name in self.active_camera_names:
      self.cameras[camera_name] = SplitRaspberryPiCamera(reactor_manager, 
        self, camera_name, self.save_dir_dict[camera_name], 
        resolution = self.resolution_dict[camera_name], 
        rotation = self.rotation_dict[camera_name], 
        reflection = self.reflection_dict[camera_name])
    self.current_camera_letter = None
    self.current_camera = None

    #Based on the GPIO.mode, define the numbers for each of the pins and
    #set the pins to their initial states.
    self.setup_GPIO_pins()

    #Set the current camera to camera A initially.
    self.set_camera(self.active_camera_names[0])

  #Set up the pins used to control the MultiCamera. This pins cannot be changed.
  #The pins are set to different numbers based on the GPIO mode.
  #The naming convention and setup for the pins comes from:
  #https://gist.github.com/skypanther/04b44eaa455f358eb70e59336c068312#file-multicam-py-L154
  #NOTE: all pins labeled b2_... to b4_... are meant for the case where
  #multiple ArduCam Multi Camera Adapters are stacked (although this is
  #possible according to the supplier of this adapter, people report errors
  #with stacking multiple adapters).
  def setup_GPIO_pins(self):
    #Set up the pin numbers based on the GPIO mode.
    if GPIO.getmode() == GPIO.BCM:
      self.channel_select_pin = 4
      self.b1_oe_pin_1 = 17
      self.b1_oe_pin_2 = 18
      self.b2_oe_pin_1 = 22
      self.b2_oe_pin_2 = 23
      self.b3_oe_pin_1 = 9
      self.b3_oe_pin_2 = 25
      self.b4_oe_pin_1 = 11
      self.b4_oe_pin_2 = 8
    elif GPIO.getmode() == GPIO.BOARD:
      self.channel_select_pin = 7
      self.b1_oe_pin_1 = 11
      self.b1_oe_pin_2 = 12
      self.b2_oe_pin_1 = 15
      self.b2_oe_pin_2 = 16
      self.b3_oe_pin_1 = 21
      self.b3_oe_pin_2 = 22
      self.b4_oe_pin_1 = 23
      self.b4_oe_pin_2 = 24
    else:
      raise Exception("Multicamera Error({0}): GPIO mode was not set to GPIO.BCM or GPIO.BOARD.".format(self))
    #Setup each GPIO pin.
    GPIO.setup(self.channel_select_pin, GPIO.OUT)
    GPIO.setup(self.b1_oe_pin_1, GPIO.OUT)
    GPIO.setup(self.b1_oe_pin_2, GPIO.OUT)
    GPIO.setup(self.b2_oe_pin_1, GPIO.OUT)
    GPIO.setup(self.b2_oe_pin_2, GPIO.OUT)
    GPIO.setup(self.b3_oe_pin_1, GPIO.OUT)
    GPIO.setup(self.b3_oe_pin_2, GPIO.OUT)
    GPIO.setup(self.b4_oe_pin_1, GPIO.OUT)
    GPIO.setup(self.b4_oe_pin_2, GPIO.OUT)
    #Initialize all of the pins to GPIO.HIGH
    GPIO.output(self.channel_select_pin, GPIO.HIGH)
    GPIO.output(self.b1_oe_pin_1, GPIO.HIGH)
    GPIO.output(self.b1_oe_pin_2, GPIO.HIGH)
    GPIO.output(self.b2_oe_pin_1, GPIO.HIGH)
    GPIO.output(self.b2_oe_pin_2, GPIO.HIGH)
    GPIO.output(self.b3_oe_pin_1, GPIO.HIGH)
    GPIO.output(self.b3_oe_pin_2, GPIO.HIGH)
    GPIO.output(self.b4_oe_pin_1, GPIO.HIGH)
    GPIO.output(self.b4_oe_pin_2, GPIO.HIGH)

  #The parameters are checked for each individual camera when they are created.
  def check_parameters(self):
    return

  #Return a string representing the multicamera.
  def __repr__(self):
    return "MultiCamera: Cameras: {0} Current Camera: {1}".format(self.cameras, self.current_camera)

  #Add the ability to compare two MultiCamera instances.
  def __eq__(self, other):
    equality_test = True
    for camera_name in self.active_camera_names:
      if camera_name in other.active_camera_names:
        equality_test = equality_test and \
          (self.cameras[camera_name] == other.cameras[camera_name])
      else:
        equality_test = False
    return equality_test

  #Allow the MultiCamera to be added as a key to sets/dictionaries.
  def __hash__(self):
    return hash(tuple(self.cameras.items()))

  #Return each the instance for each split camera managed by the multi camera.
  #Give it the letter of the camera to get the camera associated with that
  #letter.
  def get_camera(self, camera_letter):
    return self.cameras[camera_letter]

  #Given a camera letter "A", "B", "C", or "D", set the current camera to that
  #camera on the Multi Camera Adapter Board.
  def set_camera(self, camera_letter):
    self.current_camera_letter = camera_letter
    self.current_camera = self.cameras[camera_letter]
    if camera_letter == "A":
      i2c_command = ["i2cset", "-y", "1", "0x70", "0x00", "0x04"]
      self.reactor_manager.run_on_command_line(i2c_command)
      GPIO.output(self.channel_select_pin, GPIO.LOW)
      GPIO.output(self.b1_oe_pin_1, GPIO.LOW)
      GPIO.output(self.b1_oe_pin_2, GPIO.HIGH)
    elif camera_letter == "B":
      i2c_command = ["i2cset", "-y", "1", "0x70", "0x00", "0x05"]
      self.reactor_manager.run_on_command_line(i2c_command)
      GPIO.output(self.channel_select_pin, GPIO.HIGH)
      GPIO.output(self.b1_oe_pin_1, GPIO.LOW)
      GPIO.output(self.b1_oe_pin_2, GPIO.HIGH)
    elif camera_letter == "C":
      i2c_command = ["i2cset", "-y", "1", "0x70", "0x00", "0x06"]
      self.reactor_manager.run_on_command_line(i2c_command)
      GPIO.output(self.channel_select_pin, GPIO.LOW)
      GPIO.output(self.b1_oe_pin_1, GPIO.HIGH)
      GPIO.output(self.b1_oe_pin_2, GPIO.LOW)
    elif camera_letter == "D":
      i2c_command = ["i2cset", "-y", "1", "0x70", "0x00", "0x07"]
      self.reactor_manager.run_on_command_line(i2c_command)
      GPIO.output(self.channel_select_pin, GPIO.HIGH)
      GPIO.output(self.b1_oe_pin_1, GPIO.HIGH)
      GPIO.output(self.b1_oe_pin_2, GPIO.LOW)
  
  #Set the current camera to the camera specified by camera_letter, then
  #take a picture and save at the given picture_path.
  def capture_picture(self, camera_letter, picture_path):
    self.set_camera(camera_letter)
    full_command = self.current_camera.command_and_flags + ["-o", picture_path]
    self.reactor_manager.run_on_command_line(full_command)

  #Switch the camera back to A. The hardware manager will take care of
  #cleanup up the GPIO pins.
  def cleanup(self):
    self.set_camera(self.active_camera_names[0])

#This class is intended to be used with Raspberry Pi Cameras that are
#attached to an ArduCam Multi Camera 4 Channel Adapter.
# NOTE: the following is documentation for raspistill:
# https://thepihut.com/blogs/raspberry-pi-roundup/raspberry-pi-camera-board-raspistill-command-list
class SplitRaspberryPiCamera(GenericCamera):
  #Take in an instance of ReactorManager (reactor_manager), an instance
  #of MultiCamera (multi_camera), the directory to save the pictures in 
  #(save_dir), the resolution for the camera (resolution), the rotation for the 
  #camera (rotation), the reflection for the camera (reflection), and the letter 
  #of the multi camera ("A", "B", "C", or "D"), and return an instance of a 
  #SplitRaspberryPiCamera.
  #
  #NOTE: if resolution, rotation, or reflection is passed as None, then the 
  #camera will not set a specific resolution, rotation, or reflection. Instead, 
  #the camera will attempt to use the default resolution/rotation/reflection.
  def __init__(self, reactor_manager, multi_camera, camera_letter, save_dir, 
               resolution = None, rotation = None, reflection = None):
    self.multi_camera = multi_camera
    self.command_and_flags = ["raspistill", "-n", "-t", "500"]
    self.camera_letter = camera_letter
    super().__init__(reactor_manager, "SplitRaspberryPiCamera", save_dir, 
                     resolution = resolution, rotation = rotation, 
                     reflection = reflection, position = self.camera_letter)

  #Check that the camera letter is valid. I cannot add a check for the
  #resolution because it can vary with the camera.
  def check_parameters(self):
    super().check_parameters()
    if self.camera_letter not in ["A", "B", "C", "D"]:
      raise Exception("{0} Error({1}): Camera Letter {2} is not A, B, C, or D.".format(self.camera_type, self, self.camera_letter))

  #Set the resolution of the picture.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE RESOLUTION MID RUN.
  def set_initial_resolution(self, resolution):
    self.command_and_flags.append("-w")
    self.command_and_flags.append(str(resolution[0]))
    self.command_and_flags.append("-h")
    self.command_and_flags.append(str(resolution[1]))

  #Set the rotation of the picture.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE ROTATION MID RUN.
  def set_initial_rotation(self, rotation):
    if rotation != 0:
      self.command_and_flags.append("-rot")
      self.command_and_flags.append(str(rotation))
  
  #Set the reflection for the camera.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE REFLECTION MID RUN.
  def set_initial_reflection(self, reflection):
    if reflection[0]:
      self.command_and_flags.append("-hf")
    if reflection[1]:
      self.command_and_flags.append("-vf")

  #Given a file path (picture_path) for a picture, capture a picture and save 
  #it at the file path.
  #NOTE: banners are not implemented for this class.
  #NOTE: to take pictures, do not use this function, use take_picture.
  #This function is for internal use only.
  def capture_picture(self, picture_path, banner_data):
    self.multi_camera.capture_picture(self.camera_letter, picture_path)
  
  #No cleanup is needed for the specific cameras; instead all cleanup is
  #handled by the MultiCamera instance.
  def cleanup(self):
    return