from reactor_code_framework.camera.generic_camera import GenericCamera

import picamera

#See this website for setting up the hardware for PiCameras.
#https://www.arrow.com/en/research-and-events/articles/raspberry-pi-camera-options-and-usage

#This class is intended to be used with PiCameras.
#NOTE: in the past, our USB cameras have taken pictures with a resolution of
#2592x1944, but you can feel free to use other resolutions.
class RaspberryPiCamera(GenericCamera):
  #Take in a ReactorManager (reactor_manager), the directory to save the 
  #pictures in (save_dir), the camera resolution (resolution), the camera 
  #rotation (rotation), and the camera reflection (reflection), and instatiate 
  #the camera class.
  #
  #NOTE: if resolution, rotation, or reflection is passed as None, then the 
  #camera will not set a specific resolution, rotation, or reflection. Instead, 
  #the camera will attempt to use the default resolution/rotation/reflection.
  def __init__(self, reactor_manager, save_dir, resolution = None, 
               rotation = None, reflection = None):
    self.camera = picamera.PiCamera()
    super().__init__(reactor_manager, "RaspberryPiCamera", save_dir, 
                     resolution = resolution, rotation = rotation, 
                     reflection = reflection, position = None)

  #Raise an exception if the parameters aren't valid (build off of in sub 
  #class).
  def check_parameters(self):
    super().check_parameters()
    if not (self.resolution == None or \
            (self.resolution[0] <= self.camera.MAX_RESOLUTION[0] and \
             self.resolution[1] <= self.camera.MAX_RESOLUTION[1])):
      raise Exception("{0} Error({1}): Resolution {2} is not within the maximum resolution {3}.".format(self.camera_type, self, self.resolution, self.camera.MAX_RESOLUTION))
  
  #Set the resolution of the camera, using the initial resolution.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE RESOLUTION MID RUN.
  def set_initial_resolution(self, resolution):
    self.camera.resolution = resolution

  #Set the rotation of the camera, using the inital rotation.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE ROTATION MID RUN.
  def set_initial_rotation(self, rotation):
    if rotation != 0:
      self.camera.rotation = rotation
  
  #Set the reflection for the camera, using the initial reflection.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE REFLECTION MID RUN.
  def set_initial_reflection(self, reflection):
    if reflection[0]:
      self.camera.hflip = True
    if reflection[1]:
      self.camera.vflip = True
  
  #Set the auto white balance of the camera (implement in sub class).
  def set_auto_white_balance(self, auto_white_balance):
    self.camera.awb_mode = "off"
    self.camera.awb_gains = auto_white_balance

  #Given a file path (picture_path) for a picture, capture a picture and save 
  #it at the file path.
  #NOTE: banners are not implemented for this class.
  #NOTE: to take pictures, do not use this function, use take_picture.
  #This function is for internal use only.
  def capture_picture(self, picture_path, banner_data):
    self.camera.capture(picture_path)

  #Cleanup the camera.
  def cleanup(self):
    self.camera.close()