from reactor_code_framework.camera.generic_camera import GenericCamera

#You must first install fswebcam with the following instructions in order to
#use this class.

#fswebcam installation instructions (from the command line):
#>>> sudo apt-get update
#>>> sudo apt-get install fswebcam
#check that camera works: from command line, input fswebcam "image.jpg"

#Manual Page for fswebcam software:
#https://manpages.ubuntu.com/manpages/bionic/man1/fswebcam.1.html

#This class is intended to be used with USB cameras via the fswebcam program.
#NOTE: in the past, our USB cameras have taken pictures with a resolution of
#2592x1944, but you can feel free to use other resolutions.
class USBCamera(GenericCamera):
  #Take in a ReactorManager (reactor_manager), the directory to save the 
  #pictures in (save_dir), the camera resolution (resolution), the camera 
  #rotation (rotation), and the camera reflection (reflection), and instataiate 
  #the camera class.
  #
  #NOTE: if resolution, rotation, or reflection is passed as None, then the 
  #camera will not set a specific resolution, rotation, or reflection. Instead, 
  #the camera will attempt to use the default resolution/rotation/reflection.
  def __init__(self, reactor_manager, save_dir, resolution = None, 
               rotation = None, reflection = None):
    self.banner_font = '"sans:60"'
    self.command_and_flags = ["fswebcam", "--quiet"]
    super().__init__(reactor_manager, "USBCamera", save_dir, 
                     resolution = resolution, rotation = rotation, 
                     reflection = reflection, position = None)
  
  #Note, the maximum resolution of different USB cameras can vary, and I am
  #not sure of the maximum resolution of the current USB cameras. As such,
  #I cannot check the resolution here. If you recieve errors from fswebcam
  #about resolution, double check with the specifications of the hardware that
  #your resolution is not too high/low.
  def check_parameters(self):
    super().check_parameters()

  #Set the resolution for the camera.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE RESOLUTION MID RUN.
  def set_initial_resolution(self, resolution):
    self.command_and_flags.append("--resolution")
    resolution_string = "{0}x{1}".format(resolution[0], resolution[1])
    self.command_and_flags.append(resolution_string)
  
  #Set the rotation for the camera.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE ROTATION MID RUN.
  def set_initial_rotation(self, rotation):
    if rotation != 0:
      self.command_and_flags.append("--rotate")
      self.command_and_flags.append(str(rotation))
  
  #Set the reflection for the camera.
  #NOTE: THIS FUNCTION SHOULD NOT BE CALLED BY THE USER. IT IS NOT DESIGNED
  #TO CHANGE THE REFLECTION MID RUN.
  def set_initial_reflection(self, reflection):
    self.command_and_flags.append("--flip")
    flip_options = None
    if reflection[0] and reflection[1]:
      flip_options = "h,v"
    elif reflection[0] and not reflection[1]:
      flip_options = "h"
    elif not reflection[0] and reflection[1]:
      flip_options = "v"
    if flip_options != None:
      self.command_and_flags.append(flip_options)

  #Given a file path (picture_path) for a picture, capture a picture and save 
  #it at the file path. If banner_data is not None, place a banner on to the
  #bottom of each image. Use the title, sub title, and info in banner_data
  #to draw this banner.
  #NOTE: in our old reactors, the title indicated the illumination time,
  #the sub title indicated the pressure, and the info indicated the oxygen.
  #NOTE: to take pictures, do not use this function, use take_picture.
  #This function is for internal use only.
  def capture_picture(self, picture_path, banner_data):
    if banner_data == None:
      banner_flags = ["--no-banner"]
    else:
      banner_flags = ["--title", banner_data[0], 
                      "--subtitle", banner_data[1],
                      "--info", banner_data[2],
                      "--font", self.banner_font]
    full_command = self.command_and_flags + banner_flags + [picture_path]
    self.reactor_manager.run_on_command_line(full_command)

  #No cleanup is required for these cameras.
  def cleanup(self):
    return