import sys

#This class is meant to be used in all code that interfaces with hardware.
#The primary purpose of this manager is to run the cleanup code of all hardware 
#devices.
#NOTE: see the below link for the difference between GPIO.BCM and GPIO.BOARD:
#https://raspberrypi.stackexchange.com/questions/12966/what-is-the-difference-between-board-and-bcm-for-gpio-pin-numbering
class HardwareManager(object):
  #Create a list to store devices. Take in the reactor manager and store
  #it. Take in a mode to use for GPIO (either "BCM"
  #or "BOARD").
  def __init__(self, reactor_manager, GPIO_mode):
    self.reactor_manager = reactor_manager
    self.devices = []
    self.GPIO_mode = GPIO_mode
    self.check_parameters()
    self.GPIO_setup()
  
  #Check that the GPIO mode is valid.
  def check_parameters(self):
    if self.GPIO_mode not in ["BCM", "BOARD"]:
      raise Exception("Hardware Manager Error({0}): GPIO mode {1} invalid.".format(self, self.GPIO_mode))
  
  #Create a string representing all devices managed by the hardware manager.
  def __repr__(self):
    hardware_manager_string = "Hardware Manager: GPIO mode: {0}\nActive Hardware:\n".format(self.GPIO_mode)
    for device in self.devices:
      hardware_manager_string = hardware_manager_string + str(device) + "\n"
    #Remove the last new line "\n" character because it is not needed.
    hardware_manager_string = hardware_manager_string[:-1]
    return hardware_manager_string

  #Check that the hardware device has a callable cleanup method.
  def check_device(self, device):
    cleanup_fn = getattr(device, "cleanup", None)
    if not callable(cleanup_fn):
      raise Exception("Device {0} does not have a cleanup method defined. Please define this method!".format(device))

  #Add a device to the hardware manager. This should be done for every hardware
  #class that is instantiated.
  def add_device(self, device):
    self.check_device(device)
    self.devices.append(device)
  
  #Return the GPIO mode.
  def get_GPIO_mode(self):
    return self.GPIO_mode
  
  #Return True if the RPi.GPIO modules has been imported. 
  def detect_GPIO(self):
    return "RPi.GPIO" in sys.modules

  #If RPi.GPIO has been imported, then run the appropriate setup code given the 
  #user-defined GPIO mode.
  def GPIO_setup(self):
    if self.detect_GPIO():
      #This will double import RPi.GPIO, which should have no effect on the
      #code. This line is included purely to get Python to run this file. If it
      #is excluded, it will throw an error because it does not know what "GPIO"
      #is.
      import RPi.GPIO as GPIO
      #GPIO.setwarnings(False) was used in old reactor code for an unknown 
      #reason.
      if self.GPIO_mode == "BCM":
        GPIO.setmode(GPIO.BCM)
      elif self.GPIO_mode == "BOARD":
        GPIO.setmode(GPIO.BOARD)
  
  #If RPi.GPIO has been imported, run the appropriate cleanup code.
  def GPIO_cleanup(self):
    if self.detect_GPIO():
      #This will double import RPi.GPIO, which should have no effect on the
      #code. This line is included purely to get Python to run this file. If it
      #is excluded, it will throw an error because it does not know what "GPIO"
      #is.
      import RPi.GPIO as GPIO
      GPIO.cleanup()

  #Run the cleanup code for each device managed by the hardware manager.
  def cleanup(self):
    for device in self.devices:
      device.cleanup()
    self.GPIO_cleanup()