from reactor_code_framework.reactor_manager.hardware_manager import HardwareManager
from reactor_code_framework.reactor_manager.software_manager import SoftwareManager
from reactor_code_framework.reactor_manager.storage_manager import StorageManager
from reactor_code_framework.reactor_manager.io_manager import IOManager

import tkinter as tk
import tkinter.messagebox
import traceback

import sys
import subprocess

#This class manages all of the hardware devices and software classes. This 
#mostly handles the necessary setup and cleanup. Additionally, functions are
#provided for managing the disk space. These functions will cause the program
#to crash if there is not enough disk space left.
class ReactorManager(object):
  #Given the GPIO_mode ("BCM" or "BOARD"), the absolute lower limit for the
  #disk space, the low limit for the disk space for startup, and the
  #relative path of the config file, create a reactor manager.
  def __init__(self, GPIO_mode, absolute_disk_space_limit,
               startup_disk_space_limit, relative_config_file_path):
    self.cooling_warning()
    self.hardware_manager = HardwareManager(self, GPIO_mode)
    self.software_manager = SoftwareManager(self)
    self.storage_manager = StorageManager(self, absolute_disk_space_limit,
                                          startup_disk_space_limit)
    self.storage_manager.startup_check_disk_space()
    self.io_manager = IOManager(self)
    self.absolute_config_file_path = \
      self.io_manager.get_abs_path(relative_config_file_path)
    self.use_subprocess_run = False
    self.handle_python_version()
  
  #Handle different versions of python, including:
  #  -determining the function to use to run commands from the command line.
  def handle_python_version(self):
    python_version = sys.version.split(" ")[0]
    python_overall_version = int(python_version.split(".")[0])
    python_sub_version = int(python_version.split(".")[1])
    
    #This code will not work with python 2.
    if python_overall_version < 3:
      raise Exception("Python Version Error: Python 3 is required for the reactor code.")
    
    #Determine the function to use to run commands from the command line.
    if python_overall_version == 3 and python_sub_version <= 4:
      self.use_subprocess_run = False
    else:
      self.use_subprocess_run = True
  
  #Before the reactor starts, create a popup window to ask if the user
  #has turned on the water/fan in order to cool the reactor.
  def cooling_warning(self):
    yes = False
    root = tk.Tk()
    root.withdraw()
    yes = tkinter.messagebox.askyesno("Question","Is the water on?\nIs the fan on?")
    if not yes:
      raise Exception("If applicable, turn the on the water and the fan before continuing.")
    root.destroy()
  
  #Represent the Reactor Manager as a string.
  def __repr__(self):
    return "Reactor Manager: {0}\n{1}\n{2}\n{3}\nConfig File Path: {4}".format(self.hardware_manager, self.software_manager, self.storage_manager, self.io_manager, self.absolute_config_file_path)

  #Return the GPIO mode.
  def get_GPIO_mode(self):
    return self.hardware_manager.GPIO_mode

  #Add a device to the hardware manager. This should be done in the __init__
  #method of every hardware class.
  def add_hardware_device(self, device):
    self.hardware_manager.add_device(device)
  
  #Add a software element to the software manager. This should be done in the
  #__init__ method of every software class.
  def add_software_element(self, software_element):
    self.software_manager.add_software_element(software_element)
  
  #Cleanup all hardware and software managed by the hardware and software
  #managers.
  def cleanup(self):
    self.hardware_manager.cleanup()
    self.software_manager.cleanup()
  
  #In the case that the run code fails due to an exception, this function
  #can be used to print the exception (if it is called in the "except" block
  #after the "try" block that failed.)
  def print_exception(self):
    traceback.print_exc()

  #Return the remaining free disk space in gigabytes.
  def get_free_disk_space(self):
    return self.storage_manager.get_free_disk_space()

  #Check that the remaining free disk space is greater than the
  #absolute_disk_space_limit in gigabytes. Raise an exception if the remaining
  #free disk space falls below the limit.
  def check_disk_space(self):
    self.storage_manager.check_disk_space()

  #Given a path for a directory, recursively make the directory. If the
  #directory already exists, prints a message and does nothing.
  def make_directory(self, dir_path):
    self.io_manager.make_directory(dir_path)

  #Given the path to a text file, return the contents of the file as a string.
  def read_text_file(self, path):
    self.io_manager.read_text_file(path)

  #Given a path and a string contents, write the string contents to the text
  #file at the location specified by the given path.
  def write_text_file(self, path, contents):
    self.io_manager.write_text_file(path, contents)
  
  #Given a text file path, path, append the content contained in
  #additional_content to the text file. Note, if the file does not yet exist,
  #it will be created by this function.
  def write_append_text_file(self, path, additional_contents):
    self.io_manager.write_append_text_file(path, additional_contents)
  
  #Given a relative file or directory path, return the absolute path on the
  #system.
  def get_abs_path(self, relative_path):
    return self.io_manager.get_abs_path(relative_path)

  #Given a path to a file, return the directory as a string.
  def get_directory_from_path(self, path):
    return self.io_manager.get_directory_from_path(path)
  
  #Return the variables from the config file.
  def read_config_variables(self):
    return self.io_manager.read_parameters_from_yaml(self.absolute_config_file_path)
  
  #Run a command on the command line.
  def run_on_command_line(self, command_list):
    if self.use_subprocess_run:
      subprocess.run(command_list)
    else:
      subprocess.call(command_list)