import shutil

#For any class that saves data, it should refer an instance of a StorageManager
#via the ReactorManager to check if storage space is running low.
class StorageManager(object):
  #Store an instance of ReactorManager as reactor_manager. Also, this class
  #takes in the startup_disk_space_limit, which is the amount of disk space in 
  #gigabytes required for the program to start at all, and 
  #absolute_disk_space_limit. If the amount of disk space in gigabytes ever 
  #falls below absolute_disk_space_limit, then the program will raise an 
  #exception. This behavior will only occur if all sensors/cameras invoke the 
  #"check_disk_space" function through the ReactorManager every time they save 
  #data.
  def __init__(self, reactor_manager, absolute_disk_space_limit,
               startup_disk_space_limit):
    self.reactor_manager = reactor_manager
    self.absolute_disk_space_limit = absolute_disk_space_limit
    self.startup_disk_space_limit = startup_disk_space_limit
    self.check_parameters()

  #Check that the startup_disk_space_limit and disk_space_limit are valid and
  #reasonable values.
  def check_parameters(self):
    if not (isinstance(self.startup_disk_space_limit, int) or isinstance(self.startup_disk_space_limit, float)):
      raise Exception("Storage Manager Error({0}): Startup Disk Space Limit {1} is not the correct type.".format(self, self.startup_disk_space_limit))
    if not (isinstance(self.absolute_disk_space_limit, int) or isinstance(self.absolute_disk_space_limit, float)):
      raise Exception("Storage Manager Error({0}): Absolute Disk Space Limit {1} is not the correct type.".format(self, self.absolute_disk_space_limit))
    if self.absolute_disk_space_limit >= self.startup_disk_space_limit:
      raise Exception("Storage Manager Error({0}): Absolute Disk Space Limit {1} Gigabytes is greater than or equal to Startup Disk Space Limit {2} Gigabytes.".format(self, self.absolute_disk_space_limit, self.startup_disk_space_limit))
    if self.absolute_disk_space_limit < 5.0:
      raise Exception("Storage Manager Error({0}): Absolute Disk Space Limit {1} Gigabytes is less than 5.0 Gigabytes.".format(self, self.absolute_disk_space_limit))
    if self.startup_disk_space_limit < 6.5:
      raise Exception("Storage Manager Error({0}): Startup Disk Space Limit {1} Gigabytes is less than 6.5 Gigabytes.".format(self, self.startup_disk_space_limit))
  
  #Represent the storage manager as a string.
  def __repr__(self):
    return "Storage Manager: Absolute Disk Space Limit: {0} Gigabytes  Startup Disk Space Limit: {1} Gigabytes".format(self.absolute_disk_space_limit, self.startup_disk_space_limit)

  #Return the amount of storage space on the disk that is free in gigabytes.
  #NOTE: 1 GB = 1024 MB = (1024)^2 KB = (1024)^3 Bytes.
  def get_free_disk_space(self):
    return (shutil.disk_usage("/").free) / ((1024) ** 3)

  #Raise an exception if the current free disk space is less than the amount
  #of disk space required to start the program.
  def startup_check_disk_space(self):
    current_free_disk_space = self.get_free_disk_space()
    if current_free_disk_space < self.startup_disk_space_limit:
      raise Exception("Current free disk space {0} Gigabytes is less than the Startup Disk Space Limit {1} Gigabytes.".format(current_free_disk_space, self.startup_disk_space_limit))

  #Raise an exception if the current free disk space is less than the absolute
  #lower limit for disk space.
  def check_disk_space(self):
    current_free_disk_space = self.get_free_disk_space()
    if current_free_disk_space < self.absolute_disk_space_limit:
      raise Exception("Current free disk space {0} Gigabytes is less than the Absolute Disk Space Limit {1} Gigabytes.".format(current_free_disk_space, self.absolute_disk_space_limit))