import time

# NOTE: This is a software class, not a hardware class.

# The class allows the user to schedule events at different delays
# up until a certain time has been reached. For example, this class could be
# used if you wanted to have a current sensor read every second and a 
# camera take pictures every minute for 100 minutes total. The important 
# function to use is add_event (found below). Please refer to the comment above 
# this function.
#
# NOTE: the added events must not take any additional inputs. As such, it
# might be best to define these functions within your run function. This allows
# the use of "nonlocal" variables defined in the run function.
# 
# NOTE: no __eq__ or __hash__ functions are provided because instances of this
# class are not meant to be hashed or compared.
class ReactorScheduler(object):
  # This class takes in reactor_manager, an instance of ReactorManager, and
  # it takes the total time in minutes (total_time) that ReactorScheduler will 
  # run for. If total_time is None, the scheduler will run forever.
  def __init__(self, reactor_manager, total_time = None):
    self.reactor_manager = reactor_manager
    self.total_time = total_time
    self.events = []
    self.check_parameters()
    # Add this instance to the reactor manager.
    self.reactor_manager.add_software_element(self)
  
  # Check that the total_time is valid.
  def check_parameters(self):
    if not (isinstance(self.total_time, int) or isinstance(self.total_time, float) or self.total_time == None):
      raise Exception("ReactorScheduler Error({0}): Total Time {1} is not valid.".format(self, self.total_time))
    if self.total_time != None and self.total_time < 0:
      raise Exception("ReactorScheduler Error({0}): Total Time {1} is not greater than 0.".format(self, self.total_time))

  # Return a string representing the reactor scheduler. 
  def __repr__(self):
    return "ReactorScheduler: Total Time: {0}\nEvents: {2}".format(self.total_time, self.events)

  # Given an event, add the event to the scheduler.
  def add_event(self, event):
    if not isinstance(event, Event):
      raise Exception("ReactorScheduler Error({0}): Event {1} is not an instance of the Event class (or one of its subclasses).".format(self, event))
    if self.total_time != None and event.interval_time > self.total_time:
      raise Exception("ReactorScheduler Error({0}): New event interval time {1} is greater than the reactor scheduler's total time {2}.".format(self, event.interval_time, self.total_time))
    if self.total_time != None and event.start_time > self.total_time:
      raise Exception("ReactorScheduler Error({0}): New event start time {1} is greater than the reactor scheduler's total time {2}.".format(self, event.start_time, self.total_time))
    if self.total_time != None and event.end_time != None and event.end_time > self.total_time:
      raise Exception("ReactorScheduler Error({0}): New event end time {1} is greater than the reactor scheduler's total time {2}.".format(self, event.end_time, self.total_time))
    self.events.append(event)

  # Loop repeatedly, checking if the events are ready and calling them if they
  # are ready. If self.total_time is None, loop forever.
  def loop(self):
    reactor_start_time = time.time()

    no_total_time = (self.total_time == None)

    # Loop until total_time has elapsed.
    while no_total_time or \
          (time.time() - reactor_start_time <= self.total_time * 60):
      for event in self.events:
        if event.ready(reactor_start_time):
          event.call()

  # Cleanup function to be called by the reactor manager. At the moment,
  # there is nothing for this class to cleanup.
  def cleanup(self):
    return
  
class Event(object):
  # This class can be used to create events for the scheduler. 
  #
  # event_function should be a function, created in the user's run function
  #   that performs some reactor task. Note, this function should not take any
  #   inputs.
  #
  # interval_time (float) denotes the number of minutes to wait between calling
  #   this event.
  #
  # start_time (float) denotes the start time in minutes for this event. This
  #   allows the user to specify events that start a certain amount of time
  #   after the reactor has started. The default is 0, which means that the
  #   event will start running when the reactor starts.
  #
  # end_time (float) denotes the end time in minutes for this event. This
  #   allows the user to specify events that end before the reactor stops 
  #   running. The default is None, which means that the event will run until
  #   the reactor stops.
  def __init__(self, event_function, interval_time, start_time = 0, 
               end_time = None):
    self.event_function = event_function
    self.interval_time = interval_time

    self.start_time = start_time
    self.end_time = end_time

    self.last_time_event_called = None

    self.check_event_parameters()

  # Check that the inputs to the event are valid.
  def check_event_parameters(self):
    # If the event_function is not callable, raise an exception.
    if self.event_function != None and not callable(self.event_function):
      raise Exception("The event_function for an event must be callable.")

    # Check that the interval time is valid.
    if not (isinstance(self.interval_time, int) or \
            isinstance(self.interval_time, float)):
      raise Exception("The interval_time for an event must be an int or a float.")
    if self.interval_time < 0:
      raise Exception("The interval_time for an event must be greater than or equal to 0.")

    # Check the start and end times.
    if not (isinstance(self.start_time, int) or \
            isinstance(self.start_time, float)):
      raise Exception("The start_time for an event must be an int or a float.")
    if self.end_time != None and not (isinstance(self.end_time, int) or \
                                      isinstance(self.end_time, float)):
      raise Exception("The end_time for an event must be an int or a float.")
    if self.end_time != None and self.start_time >= self.end_time:
      raise Exception("The start_time for an event must be less than the end_time.")

  # This function returns True if the event is ready to be called, and False
  # otherwise.
  #
  # reactor_start_time is the time that the reactor started at in seconds.
  #
  # NOTE: Current time is in seconds, so the start/end/interval times must be 
  # converted to seconds from minutes.
  def ready(self, reactor_start_time):
    current_time = time.time()
    # Check that the start time has been surpassed.
    if current_time - reactor_start_time < self.start_time * 60:
      return False
    
    # Check that the end time has not been reached.
    if (self.end_time != None) and \
       (current_time - reactor_start_time >= self.end_time * 60):
      return False
    
    # Check that the interval time has elapsed since the last time this event
    # was called.
    if (self.last_time_event_called != None) and \
       ((current_time - self.last_time_event_called) < self.interval_time * 60):
      return False
    
    return True

  # Store the current time as the last_time_event_called (in seconds), then call
  # the event function.
  def call(self):
    self.last_time_event_called = time.time()
    self.event_function()

class SequenceEvent(Event):
  # This class is intended to be used for events with multiple functionalities
  # where the user wants to add a delay between functionalities. For example,
  # if the user wants to turn off stiring, then wait 0.5 minutes, then take
  # a picture, this class could be used.
  #
  # event_functions should be a list of functions that the user wants to be
  #   called in succession.
  #
  # in_between_delays (int or Float List): a list of delays in minutes between 
  #   the event_functions. This list should have a length that is one less than
  #   the event_functions. Then, the 0th element in in_between_delays 
  #   corresponds to the delay time between the 0th and 1st functions in 
  #   event_functions (the 1st element in in_between_delays corresponds to the 
  #   delay time between the 1st and 2nd functions in event_functions, and so 
  #   on).
  #
  # interval_time (float) denotes the number of minutes to wait between calling
  #   this event. 
  #   NOTE: sum of in_between_delays should not be greater than interval_time.
  #
  # start_time (float) denotes the start time in minutes for this event. This
  #   allows the user to specify events that start a certain amount of time
  #   after the reactor has started. The default is 0, which means that the
  #   event will start running when the reactor starts.
  #
  # end_time (float) denotes the end time in minutes for this event. This
  #   allows the user to specify events that end before the reactor stops 
  #   running. The default is None, which means that the event will run until
  #   the reactor stops.
  #   NOTE: if an end_time is specified, the sequence event may continue 
  #   slightly past the end_time in order to finish the current sequence.
  def __init__(self, event_functions, in_between_delays, interval_time, 
               start_time = 0, end_time = None):
    self.event_functions = event_functions
    self.in_between_delays = in_between_delays

    # Combine the interval_time and the in_between_delays for use in ready.
    # (the interval_time can be treated as the "in between delay" for the 0th 
    # function).
    self.interval_time_and_delays = [interval_time] + in_between_delays

    self.num_functions = len(event_functions)
    self.next_to_call_index = 0

    super().__init__(None, interval_time, start_time = start_time, 
                     end_time = end_time)
    
    # We need an last time called for the functions within the event, as well
    # as the last_time_event_called for the event.
    self.last_time_subfunction_called = None
  
  # Check that the inputs to the event are valid.
  def check_event_parameters(self):
    super().check_event_parameters()

    # Check that every function in self.event_functions is callable.
    for event_function in self.event_functions:
      if not callable(event_function):
        raise Exception("Every function in event_functions for a sequence event must be callable.")

    # Check that the in_between_delays are valid.
    for delay in self.in_between_delays:
      if not (isinstance(delay, int) or isinstance(delay, float)):
        raise Exception("Every delay time in in_between_delays for a sequence event must be an int or a float.")
      if delay < 0:
        raise Exception("Every delay time in in_between_delays for a sequence event must be greater than or equal to 0.")

    # Check that the sum of in_between_delays is not greater than the 
    # interval_time.
    if sum(self.in_between_delays) > self.interval_time:
      raise Exception("Sum of in_between_delays for a sequence event must be less than or equal to the interval_time.")

  # This function returns True if the event is ready to be called, and False
  # otherwise.
  #
  # reactor_start_time is the time that the reactor started at in seconds.
  #
  # For the 0th function in the sequence, ready will check that the
  # interval_time has elapsed (relative to the last time the event was called). 
  # Otherwise, ready will check the appropriate delay time (relative to the
  # last time a subfunction was called).
  #
  # NOTE: Current time is in seconds, so the start/end/interval times must be 
  # converted to seconds from minutes.
  def ready(self, reactor_start_time):
    current_time = time.time()
    # Check that the start time has been surpassed.
    if current_time - reactor_start_time < self.start_time * 60:
      return False
    
    # Check that the end time has not been reached.
    # Only allow the end time to stop the sequence event if we are waiting
    # to call the 0th function in the sequence (never stop due to end time in 
    # the middle of a sequence).
    if (self.next_to_call_index == 0) and (self.end_time != None) and \
       (current_time - reactor_start_time >= self.end_time * 60):
      return False
    
    # If the sequence event is on the 0th sub-event, compare to the last time
    # the entire event was called.
    if self.next_to_call_index == 0:
      last_time = self.last_time_event_called
    # Otherwise, compare to the last time a sub-function in the event was
    # called.
    # NOTE: self.last_time_subfunction_called will never be None when this case
    # is reached, since a sequence event always calls its 0th function first.
    else:
      last_time = self.last_time_subfunction_called

    if (last_time != None) and ((current_time - last_time) < 
      self.interval_time_and_delays[self.next_to_call_index] * 60):
        return False 
    
    return True

  # Store the current time as the last_time_subfunction_called (in seconds).
  # If the next subfunction to be called is the 0th subfunction, also store
  # this current time as the last_time_event_called (in seconds). Call the
  # next function to be called. Finally, update the next event function to be 
  # called (increment by 1, wrapping around to the beginning).
  def call(self):
    current_time = time.time()
    self.last_time_subfunction_called = current_time
    if self.next_to_call_index == 0:
      self.last_time_event_called = current_time

    self.event_functions[self.next_to_call_index]()
    
    self.next_to_call_index = (self.next_to_call_index + 1) % self.num_functions