from reactor_code_framework.relay.generic_relay import GenericRelay

import smbus2

# This class is intended to be used with the DockerPi 4 Channel Stackable Relay. 
# This does not use GPIO pins; instead it uses i2c. This code requires the 
# "smbus" python module. Documentation for this relay can be found at the 
# following link:
# https://wiki.52pi.com/index.php?title=EP-0099
class DockerStackableRelay(GenericRelay):
  # Store the dip switch values and the relay number. Also, store the bus 
  # number, which cannot be changed. Then, call the super().__init__ to complete 
  # the generic relay setup.
  # NOTE: The dip switch values should be "on" or "off", and you can find/change
  # these switches on the physical relay board. For more details, see
  # https://wiki.52pi.com/index.php?title=EP-0099.
  def __init__(self, reactor_manager, dip_switch_1, dip_switch_2, relay_num):
    self.dip_switch_1 = dip_switch_1
    self.dip_switch_2 = dip_switch_2
    self.relay_num = relay_num
    self.bus_num = 1
    self.bus = smbus2.SMBus(self.bus_num)
    super().__init__(reactor_manager, "DockerStackableRelay", 
                     (self.dip_switch_1, self.dip_switch_2, self.relay_num))
  
  # Check the dip switch and relay number values.
  def check_parameters(self):
    # Check that the dip switch 1 and 2 are either "on" or "off".
    if self.dip_switch_1 not in ["on", "off"]:
      raise Exception("Relay Error({0}): dip switch 1 setting {1} is not valid.".format(self, self.dip_switch_1))
    if self.dip_switch_2 not in ["on", "off"]:
      raise Exception("Relay Error({0}): dip switch 2 setting {1} is not valid.".format(self, self.dip_switch_2))
    # Check that the relay number is between 1 and 4 (inclusive on both ends).
    if not (isinstance(self.relay_num, int) and 1 <= self.relay_num <= 4):
      raise Exception("Relay Error({0}): relay number {1} not valid.".format(self, self.relay_num))

  # Return a string representing the relay.
  def __repr__(self):
    return "{0}: Dip Switch 1: {1}  Dip Switch 2: {2}  Relay Number: {3}  State: {4}".format(self.relay_type, self.dip_switch_1, self.dip_switch_2, self.relay_num, self.state)
  
  # Return the i2c address based on the dip switch settings.
  def get_i2c_address(self):
    if self.dip_switch_1 == "off" and self.dip_switch_2 == "off":
      i2c_address = 0x10
    elif self.dip_switch_1 == "off" and self.dip_switch_2 == "on":
      i2c_address = 0x11
    elif self.dip_switch_1 == "on" and self.dip_switch_2 == "off":
      i2c_address = 0x12
    elif self.dip_switch_1 == "on" and self.dip_switch_2 == "on":
      i2c_address = 0x13
    return i2c_address

  # Change the registers at the i2c address to control the relays.
  def set_relay_state(self, state):
    i2c_addr = self.get_i2c_address()
    open_value = 0x00
    close_value = 0xFF

    if state == "open":
      self.bus.write_byte_data(i2c_addr, self.relay_num, open_value)
    elif state == "close":
      self.bus.write_byte_data(i2c_addr, self.relay_num, close_value)