def generate_generic_relay_error_string(function_name):
  return "Generic Relay Instance: DO NOT USE! " + \
         "Please override {0} in the ".format(function_name) + \
         "specific relay type's sub class."

#Generic Relay Class. Not intended for use in projects. Use specific subclasses
#instead.
class GenericRelay(object):
  #Construct the Relay based on the type and position. Set the relay to open
  #initially. reactor_manager should be an instance of ReactorManager
  #relay_type is a string that represents the type of the relay.
  #position is a variable that contains all the information about the way 
  #to access the relay. __init__ needs to be defined in the sub class.
  def __init__(self, reactor_manager, relay_type, position):
    self.reactor_manager = reactor_manager
    self.relay_type = relay_type
    self.position = position
    self.state = None
    self.check_parameters()
    self.open()
    #Pass the instance for the device to the reactor manager.
    self.reactor_manager.add_hardware_device(self)

  #Raises an exception if the user input Relay parameters are 
  #not of the proper form. This needs to be overidden in the sub class.
  def check_parameters(self):
    raise Exception(generate_generic_relay_error_string("check_parameters"))

  #Determines how Python represents the relay as a string (and how it prints).
  #This needs to be overidden in the sub class.
  def __repr__(self):
    raise Exception(generate_generic_relay_error_string("__repr__"))
  
  #Allows for the comparison of two Relay instances.
  def __eq__(self, other):
    return self.relay_type == other.relay_type and \
           self.position == other.position and \
           self.state == other.state
  
  #Allows for the use of Relay instances in sets and as keys in dictionaries.
  def __hash__(self):
    return hash(self.relay_type, self.position)

  #Takes either "open" or "close" as the state input and interacts with the
  #hardware (using the proper python library) to open or close the relay
  #respectively. This needs to be overidden in the sub class.
  def set_relay_state(self, state):
    raise Exception(generate_generic_relay_error_string("set_relay_state"))
  
  #Open the relay (i.e. prevent current from running through the relay).
  def open(self):
    new_state = "open"
    self.set_relay_state(new_state)
    self.state = new_state

  #Close the relay (i.e. allow current to run through the relay).
  def close(self):
    new_state = "close"
    self.set_relay_state(new_state)
    self.state = new_state
  
  #Return the current state of the relay.
  def get_state(self):
    return self.state
  
  #Call this function to "clean-up" the relay in a try-except at the end of the
  #code. In case the code crashes, this ensures the relays are set back to
  #the default state.
  def cleanup(self):
    self.open()